/*********************************************************************
*               SEGGER MICROCONTROLLER SYSTEME GmbH                  *
*       Solutions for real time microcontroller applications         *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2007  SEGGER Microcontroller Systeme GmbH         *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS version: 3.52d                                         *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_M52233.c                     for ColdFire CPU

Purpose : Initializes and handles the hardware for the OS as far
          as required by the OS.
          Feel free to modify this file acc. to your
          target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"

/*********************************************************************
*
*       Configuration
*
*********************************************************************/

#define ALLOW_NESTED_INTERRUPTS 1   // Allow nested interrupts

/*********************************************************************
*
*       Clock frequency settings
*/
#ifndef   OS_FSYS
  #define OS_FSYS   (60000000)        /* Default, may depend on PLL */
#endif

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ  (1000)
#endif

#define OS_TIMER_PRESCALE  (2)        /* prescaler for system timer */

/*********************************************************************
*
*       UART settings for OSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to OSView, please define it to be -1
*       Currently the standard code enables UART 0 per default
*/
#ifndef   OS_UART
  #define OS_UART (0)
#endif

#ifndef   OS_BAUDRATE
  #define OS_BAUDRATE (38400)
#endif

#define BAUDRATEREG  ((OS_FSYS/1000*125) / (OS_BAUDRATE * 4))
#if BAUDRATEREG > 0xFFFF
  #error Cannot set up UART for the OS_FSYS and OS_BAUDRATE.
#endif
#if BAUDRATEREG < 0x02
  #error Cannot set up UART for the OS_FSYS and OS_BAUDRATE.
#endif

/****** End of configuration settings *******************************/

#define OS_UART_USED (OS_UART == 0)

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
#define _NUM_INT_SOURCES          64

/* SFRs definition for PIT0 */
#define PIT0_PCSR                 (*(volatile unsigned short*) 0x40150000)
#define PIT0_PMR                  (*(volatile unsigned short*) 0x40150002)
#define PIT0_CNTR                 (*(volatile unsigned short*) 0x40150004)
#define PIT_PCSR_PIF              0x0004

/* SFRs definition for PLL */
#define MCF_CLOCK_SYNCR           (*(volatile unsigned short*) 0x40120000)
#define MCF_CLOCK_SYNSR           (*(volatile unsigned char*)  0x40120002)
#define MCF_CLOCK_CCHR            (*(volatile unsigned char*)  0x40120008)

/* Bit definitions and macros for MCF_CLOCK_SYNCR */
#define MCF_CLOCK_SYNCR_PLLEN     (0x0001)
#define MCF_CLOCK_SYNCR_PLLMODE   (0x0002)
#define MCF_CLOCK_SYNCR_CLKSRC    (0x0004)
#define MCF_CLOCK_SYNCR_LOCRE     (0x0800)
#define MCF_CLOCK_SYNCR_MFD(x)    (((x)&0x0007)<<12)

/* Bit definitions and macros for MCF_CLOCK_SYNSR */
#define MCF_CLOCK_SYNSR_LOCK      (0x08)

/* SFR definitions for interrupt controller */
#define ICR_BASE_ADDR  0x40000C40
#define IMRH           (*(volatile unsigned long*)0x40000C08)
#define IMRL           (*(volatile unsigned long*)0x40000C0C)


#if OS_UART_USED
/*  UART register definitions */

#define UMR1_0                 (*(volatile unsigned char*)0x40000200)
#define UMR2_0                 (*(volatile unsigned char*)0x40000200)
#define USR_0                  (*(volatile unsigned char*)0x40000204)
#define UCSR_0                 (*(volatile unsigned char*)0x40000240)
#define UCR_0                  (*(volatile unsigned char*)0x40000208)
#define URB_0                  (*(volatile unsigned char*)0x4000020C)
#define UTB_0                  (*(volatile unsigned char*)0x4000020C)
#define UIPCR_0                (*(volatile unsigned char*)0x40000210)
#define UACR_0                 (*(volatile unsigned char*)0x40000210)
#define UISR_0                 (*(volatile unsigned char*)0x40000214)
#define UIMR_0                 (*(volatile unsigned char*)0x40000214)
#define UIP_0                  (*(volatile unsigned char*)0x40000234)
#define UBG0_0                 (*(volatile unsigned char*)0x40000218)
#define UBG1_0                 (*(volatile unsigned char*)0x4000021C)
#define UOP0_0                 (*(volatile unsigned char*)0x40000238)
#define UOP1_0                 (*(volatile unsigned char*)0x40000238)
#define UGPIO                  (*(volatile unsigned char*)0x40100071)
#endif

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
static void _OS_IrqDefaultHandler(void);

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#ifndef    OS_ISRHANDLERTABLE_ROM
  #define  OS_ISRHANDLERTABLE_ROM 0
#endif

#if (OS_ISRHANDLERTABLE_ROM != 0)
static OS_ISR_HANDLER* const _papISR[_NUM_INT_SOURCES] = {
  _OS_IrqDefaultHandler,      // Handler for interrupt 0
  _OS_IrqDefaultHandler,      // Handler for interrupt 1
  _OS_IrqDefaultHandler,      // Handler for interrupt 2
  _OS_IrqDefaultHandler,      // Handler for interrupt 3
  _OS_IrqDefaultHandler,      // Handler for interrupt 4
  _OS_IrqDefaultHandler,      // Handler for interrupt 5
  _OS_IrqDefaultHandler,      // Handler for interrupt 6
  _OS_IrqDefaultHandler,      // Handler for interrupt 7
  _OS_IrqDefaultHandler,      // Handler for interrupt 8
  _OS_IrqDefaultHandler,      // Handler for interrupt 9
  _OS_IrqDefaultHandler,      // Handler for interrupt 10
  _OS_IrqDefaultHandler,      // Handler for interrupt 11
  _OS_IrqDefaultHandler,      // Handler for interrupt 12
  _OS_IrqDefaultHandler,      // Handler for interrupt 13
  _OS_IrqDefaultHandler,      // Handler for interrupt 14
  _OS_IrqDefaultHandler,      // Handler for interrupt 15
  _OS_IrqDefaultHandler,      // Handler for interrupt 16
  _OS_IrqDefaultHandler,      // Handler for interrupt 17
  _OS_IrqDefaultHandler,      // Handler for interrupt 18
  _OS_IrqDefaultHandler,      // Handler for interrupt 19
  _OS_IrqDefaultHandler,      // Handler for interrupt 20
  _OS_IrqDefaultHandler,      // Handler for interrupt 21
  _OS_IrqDefaultHandler,      // Handler for interrupt 22
  _OS_IrqDefaultHandler,      // Handler for interrupt 23
  _OS_IrqDefaultHandler,      // Handler for interrupt 24
  _OS_IrqDefaultHandler,      // Handler for interrupt 25
  _OS_IrqDefaultHandler,      // Handler for interrupt 26
  _OS_IrqDefaultHandler,      // Handler for interrupt 27
  _OS_IrqDefaultHandler,      // Handler for interrupt 28
  _OS_IrqDefaultHandler,      // Handler for interrupt 29
  _OS_IrqDefaultHandler,      // Handler for interrupt 30
  _OS_IrqDefaultHandler,      // Handler for interrupt 31
  _OS_IrqDefaultHandler,      // Handler for interrupt 32
  _OS_IrqDefaultHandler,      // Handler for interrupt 33
  _OS_IrqDefaultHandler,      // Handler for interrupt 34
  _OS_IrqDefaultHandler,      // Handler for interrupt 35
  _OS_IrqDefaultHandler,      // Handler for interrupt 36
  _OS_IrqDefaultHandler,      // Handler for interrupt 37
  _OS_IrqDefaultHandler,      // Handler for interrupt 38
  _OS_IrqDefaultHandler,      // Handler for interrupt 39
  _OS_IrqDefaultHandler,      // Handler for interrupt 40
  _OS_IrqDefaultHandler,      // Handler for interrupt 41
  _OS_IrqDefaultHandler,      // Handler for interrupt 42
  _OS_IrqDefaultHandler,      // Handler for interrupt 43
  _OS_IrqDefaultHandler,      // Handler for interrupt 44
  _OS_IrqDefaultHandler,      // Handler for interrupt 45
  _OS_IrqDefaultHandler,      // Handler for interrupt 46
  _OS_IrqDefaultHandler,      // Handler for interrupt 47
  _OS_IrqDefaultHandler,      // Handler for interrupt 48
  _OS_IrqDefaultHandler,      // Handler for interrupt 49
  _OS_IrqDefaultHandler,      // Handler for interrupt 50
  _OS_IrqDefaultHandler,      // Handler for interrupt 51
  _OS_IrqDefaultHandler,      // Handler for interrupt 52
  _OS_IrqDefaultHandler,      // Handler for interrupt 53
  _OS_IrqDefaultHandler,      // Handler for interrupt 54
  _OS_IrqDefaultHandler,      // Handler for interrupt 55
  _OS_IrqDefaultHandler,      // Handler for interrupt 56
  _OS_IrqDefaultHandler,      // Handler for interrupt 57
  _OS_IrqDefaultHandler,      // Handler for interrupt 58
  _OS_IrqDefaultHandler,      // Handler for interrupt 59
  _OS_IrqDefaultHandler,      // Handler for interrupt 60
  _OS_IrqDefaultHandler,      // Handler for interrupt 61
  _OS_IrqDefaultHandler,      // Handler for interrupt 62
  _OS_IrqDefaultHandler,      // Handler for interrupt 63
};

#else
static OS_ISR_HANDLER* _papISR[_NUM_INT_SOURCES];
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _ISR_TickHandler()
*       the OS system interrupt, handles OS timer
*/
static void _ISR_TickHandler(void) {
  PIT0_PCSR |= PIT_PCSR_PIF;       /* reset Timer interrupt flag */
  OS_HandleTick();
}

/*********************************************************************
*
*       OS_IrqDefaultHandler
*
*/
static void _OS_IrqDefaultHandler(void) {
#if OS_DEBUG
  OS_DI();
  while (1) {
  //
  // The NOPs are necessary for bdm debugger,
  // otherwise the bdm debugger locks up
  //
  #ifdef __IAR_SYSTEMS_ICC__
    __asm("NOP");
    __asm("NOP");
  #else
    asm(NOP);
    asm(NOP);
  #endif
  }
#endif
}

#if (OS_ISRHANDLERTABLE_ROM == 0)
  #define _INIT_VECTORS() _InitVectors()

/*********************************************************************
*
*       _InitVectors()
*/

static void _InitVectors(void) {
  int i;
  OS_ISR_HANDLER* pIsrHandler;

  pIsrHandler = _OS_IrqDefaultHandler;
  for (i = 0; i < _NUM_INT_SOURCES; +i++) {
    _papISR[i] = pIsrHandler;
  }
}
#else
  #define _INIT_VECTORS()
#endif

#if OS_UART_USED
/*********************************************************************
*
*       _Uart0Handler(void)
*/
static void _Uart0Handler(void) {
  int c;
  if (USR_0 & 0x01) {                  // Rx interrupt
    c = URB_0;
    if ((USR_0 & 0xF0) == 0x00) {      // check if errors occured
      OS_OnRx(c);
    } else {
      USR_0 &= 0x0F;                   // clear error bits
    }
  }
  if (UISR_0 & 0x01) {                 // Tx interrupt
    if (OS_OnTx()) {                   // No more characters to send?
      UIMR_0 = 0x02;                   // Disable Tx interrupt (Rx interrupt is still enabled)
    }
  }
}
#endif

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
#define OS_TIMER_RELOAD (OS_FSYS/OS_TIMER_PRESCALE/OS_TICK_FREQ)
#if (OS_TIMER_RELOAD >= 0x10000)
  #error "Timer 1 can not be used, please check configuration"
#endif

void OS_InitHW(void) {
  OS_DI();
  _INIT_VECTORS();
  //
  // Init Pll
  //
  MCF_CLOCK_CCHR  = 0x05;
  MCF_CLOCK_SYNCR = MCF_CLOCK_SYNCR_MFD(4) | MCF_CLOCK_SYNCR_CLKSRC| MCF_CLOCK_SYNCR_PLLMODE | MCF_CLOCK_SYNCR_PLLEN;
  while ((MCF_CLOCK_SYNSR & MCF_CLOCK_SYNSR_LOCK) == 0);
  //
  // Init PIT0 to generate periodic interrupt = 1ms
  //
  OS_EnableISR(55, _ISR_TickHandler, 1, 0);
  PIT0_PCSR  = 0x0000;          /* prescaler 2^0=1             */
  PIT0_PMR   = OS_TIMER_RELOAD; /* reload value for timer      */
  PIT0_PCSR |=  (1 << 0)        /* 1 = enabled EN              */
              | (1 << 1)        /* 1 = counter reload RLD      */
              | (0 << 2)        /* int reset bit PIF           */
              | (1 << 3)        /* overwrite bit PIE           */
              | (1 << 4)        /* overwrite bit OVW           */
              | (0 << 5)        /* debug mode                  */
              | (0 << 6)        /* doze mode bit               */
              | (0 << 7);       /* reserved                    */
  OS_COM_Init();
  OS_RestoreI();
}

/*********************************************************************
*
*       Idle loop  (OS_Idle)
*
*       Please note:
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggeling an output or incrementing a counter)
*/
void OS_Idle(void) {         /* Idle loop: No task is ready to exec */
  while (1) {
  //
  // The NOPs are necessary for bdm debugger,
  // otherwise the bdm debugger locks up
  //
  #ifdef __IAR_SYSTEMS_ICC__
    __asm("NOP");
    __asm("NOP");
  #else
    asm(NOP);
    asm(NOP);
  #endif
  }
}

/*********************************************************************
*
*       Get time [cycles]
*
*       This routine is required for task-info via OSView or high
*       resolution time maesurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  unsigned int t_cnt;
  OS_U32 time;
  t_cnt = OS_TIMER_RELOAD-PIT0_CNTR+1;             /* Read current timer value   */
  time  = OS_GetTime32();                          /* Read current OS time       */
  if ((PIT0_PCSR & PIT_PCSR_PIF) == PIT_PCSR_PIF) { /* Timer Interrupt pending ?  */
    time++;                                        /* Adjust result              */
    t_cnt = OS_TIMER_RELOAD-PIT0_CNTR+1;           /* Read current timer value   */
  }
  return (OS_TIMER_RELOAD * time) + t_cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us
*
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagonstics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  Cycles *= OS_TIMER_PRESCALE;   /* eliminate effect of prescaler which is fixed to OS_TIMER_PRESCALE */
  return Cycles/(OS_FSYS/1000000);
}

/*********************************************************************
*
*       Communication for OSView via UART (optional)
*
**********************************************************************
*/
#if OS_UART_USED

/*********************************************************************
*
*       OS_COM_Send1()
*/
void OS_COM_Send1(OS_U8 c) {
  UTB_0  = c;                          // store character in tx register
  UIMR_0 = 0x03;                       // enable Tx and Rx interrupt
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize UART for OSView
*/
void OS_COM_Init(void) {            /* Initialize UART, enable UART interrupts */
  OS_EnableISR(13, _Uart0Handler, 1, 1);
  UGPIO  = 0x05;                       // Init GPIO pins.
  UCR_0  = 0x30;                       // Reset transmitter
  UCR_0  = 0x20;                       // Reset reciver
  UCR_0  = 0x10;                       // Reset Mode Register
  UMR1_0 = 0x13;                       // 8N1
  UMR2_0 = 0x07;
  UCSR_0 = 0xDD;                       // Set Rx and Tx baud by SYSTEM CLOCK
  UIMR_0 = 0x02;                       // Enable Rx interrupts
  UBG0_0 = (OS_U8)(BAUDRATEREG >> 8);  // Set baud settings
  UBG1_0 = (OS_U8)BAUDRATEREG;        
  UCR_0  = 0x05;                       // Enable receiver and transmitter
}

#else  /* selected UART not supported, using dummies */

void OS_COM_Init(void) {}
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           /* avoid compiler warning */
  OS_COM_ClearTxActive();  /* let the OS know that Tx is not busy */
}

#endif /*  OS_UART_USED  */

/****** Final check of configuration ********************************/

#ifndef OS_UART_USED
  #error "OS_UART_USED has to be defined"
#endif

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler
*
*       Detect reason for IRQ and call correspondig service routine.
*       OS_irq_handler is called from OS_IRQ_SERVICE function
*       found in RTOSVect.asm
*/
void OS_irq_handler(int Status) {
  int ISRIndex;
  int OS_Ipl_Prev;
  OS_Ipl_Prev = OS_Ipl_EI;
  ISRIndex = ((Status >> 18) & 0xFF) - 64;                  // Extract ISR Index
  OS_Ipl_EI = (*(char*)(ICR_BASE_ADDR + ISRIndex)) >> 3;    // Get IPL 
#if OS_DEBUG
  if ((ISRIndex >= _NUM_INT_SOURCES) || (ISRIndex == 0)) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
#if ALLOW_NESTED_INTERRUPTS
  OS_EnterNestableInterrupt();         // Now interrupts may be reenabled. If nesting should be allowed
#else
  OS_EnterInterrupt();                 // Inform OS that interrupt handler is running
#endif
  _papISR[ISRIndex]();                 // Call installed handler
#if ALLOW_NESTED_INTERRUPTS
  OS_LeaveNestableInterrupt();         // Leave nestable interrupt, perform task switch if required
#else
  OS_LeaveInterrupt();                 // Leave interrupt, perform task switch if required
#endif
  OS_Ipl_EI = OS_Ipl_Prev;
}

/*********************************************************************
*
*       OS_EnableISR
*/
OS_ISR_HANDLER* OS_EnableISR(int ISRIndex, OS_ISR_HANDLER* pISRHandler, int Prio, int SubPrio) {
  OS_ISR_HANDLER*  pOldHandler;
  int i;

#if OS_DEBUG
  if ((ISRIndex >= _NUM_INT_SOURCES) || (ISRIndex == 0) || (Prio > OS_Ipl_DI)) {
    OS_Error(OS_ERR_ISR_INDEX);
    return NULL;
  }
#endif
  OS_DI();
  pOldHandler = _papISR[ISRIndex];       // Get old handler

#if (OS_ISRHANDLERTABLE_ROM == 0)
  _papISR[ISRIndex] = pISRHandler;       // Install new handler
#endif
  //
  // Check if the interrupt level/priority is unique
  //
  for (i = 8; i < _NUM_INT_SOURCES; i++) {
    if (*(char*) (ICR_BASE_ADDR + i) == ((Prio << 3) | SubPrio)) {
      OS_Error(OS_ERR_ISR_VECTOR);
      return NULL;
    }
  }
  *(char*)(ICR_BASE_ADDR + ISRIndex) = (Prio << 3) | SubPrio;   // Set IPL register
  if ( ISRIndex < 32) {                   // Set interrupt mask register IMRL, IMRH
    IMRL &= ~((1 << ISRIndex) | 0x0001);
  } else {
    IMRL &= ~(0x0001);                    // IMRL.bit0 has to be reset to enable any interrupts
    IMRH &= ~(1 << (ISRIndex-32));
  }
  OS_RestoreI();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_DisableISR
*/
void OS_DisableISR(int ISRIndex) {
#if OS_DEBUG
  if (ISRIndex >= _NUM_INT_SOURCES || (ISRIndex == 0)) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
  OS_DI();
  if ( ISRIndex < 32) {                   // Reset interrupt mask register
    IMRL |= (1 << ISRIndex);
  } else {
    IMRH |= (1 << (ISRIndex - 32));
  }
  *(char*)(ICR_BASE_ADDR + ISRIndex) = 0x00;   // Set IPL register to 0
#if (OS_ISRHANDLERTABLE_ROM == 0)
  _papISR[ISRIndex] = _OS_IrqDefaultHandler;   // Set default handler
#endif
  OS_RestoreI();
}

#ifndef __IAR_SYSTEMS_ICC__
/*********************************************************************
*
*       OS_SegmentInit_CodeWarrior
*
*       This routine is only for Metrowerks CodeWarrior IDE
*/
void OS_SegmentInit_CodeWarrior(void);
void OS_SegmentInit_CodeWarrior(void) {
  extern char __COPY_START[];
  extern char __COPY_DST[];
  extern char __COPY_END[];
  extern char __ZERO_START[];
  extern char __ZERO_END[];
  register unsigned int n;
  register unsigned char *dp, *sp;

  /*
   * Move initialized data from ROM to RAM.
   */
  if (__COPY_START != __COPY_DST) {
    dp = (far unsigned char *)__COPY_DST;
    sp = (unsigned char *)__COPY_START;
    n = (unsigned int)(__COPY_END - __COPY_DST);
    while (n--) {
      *dp++ = *sp++;
    }
  }

  /*
   * Zero uninitialized data
   */
  if (__ZERO_START != __ZERO_END)  {
    sp = (unsigned char *)__ZERO_START;
    n = (unsigned int)(__ZERO_END - __ZERO_START);
    while (n--) {
      *sp++ = 0;
    }
  }
}
#endif

/*****  EOF  ********************************************************/

